@@ -56,7 +56,7 @@ typedef struct
         /* full results are only possible for the rendered current page */
         int current_page;
         GArray *current_page_results;
-        guchar *other_page_flags; /* length n_pages + 1, first element ignored */
+        int *other_page_flags; /* length n_pages + 1, first element ignored */
         int start_page;   /* skip this one as we iterate, since we did it first */
         int search_page;  /* the page we're searching now */
         TextOutputDev *output_dev;
@@ -399,61 +399,69 @@ pdf_document_render (EvDocument  *document,
 					   draw.width, draw.height);
 }
 
-static void
-pdf_document_search_emit_found (PdfDocumentSearch *search)
+double
+pdf_document_find_get_progress (EvDocumentFind *document_find)
 {
-        PdfDocument *pdf_document = search->document;
-        int n_pages;
-        int pages_done;
-        GArray *tmp_results;
-        int i;
+	PdfDocumentSearch *search = PDF_DOCUMENT (document_find)->search;
+	int n_pages, pages_done;
+ 
+	n_pages = ev_document_get_n_pages (EV_DOCUMENT (document_find));
+	if (search->search_page > search->start_page) {
+		pages_done = search->search_page - search->start_page;
+	} else if (search->search_page == search->start_page) {
+		pages_done = n_pages;
+	} else {
+		pages_done = n_pages - search->start_page + search->search_page;
+	}
 
-        n_pages = ev_document_get_n_pages (EV_DOCUMENT (search->document));
-        if (search->search_page > search->start_page) {
-                pages_done = search->search_page - search->start_page;
-        } else if (search->search_page == search->start_page) {
-                pages_done = n_pages;
-        } else {
-                pages_done = n_pages - search->start_page + search->search_page;
-        }
+	return pages_done / (double) n_pages;
+}
 
-        tmp_results = g_array_new (FALSE, FALSE, sizeof (EvFindResult));
-        g_array_append_vals (tmp_results,
-                             search->current_page_results->data,
-                             search->current_page_results->len);
+int
+pdf_document_find_page_has_results (EvDocumentFind *document_find,
+				    int             page)
+{
+	PdfDocumentSearch *search = PDF_DOCUMENT (document_find)->search;
 
-        /* Now append a bogus element for each page that has a result in it,
-         * that is not the current page
-         */
-        i = 1;
-        while (i <= n_pages) {
-                if (i != pdf_document->page &&
-                    search->other_page_flags[i]) {
-                        EvFindResult result;
+	g_return_val_if_fail (search != NULL, FALSE);
 
-                        result.page_num = i;
+	return search->other_page_flags[page];
+}
 
-                        /* Use bogus coordinates, again we can't get coordinates
-                         * until this is the current page because TextOutputDev
-                         * isn't good enough
-                         */
-                        result.highlight_area.x = -1;
-                        result.highlight_area.y = -1;
-                        result.highlight_area.width = 1;
-                        result.highlight_area.height = 1;
+int
+pdf_document_find_get_n_results (EvDocumentFind *document_find)
+{
+	PdfDocumentSearch *search = PDF_DOCUMENT (document_find)->search;
 
-                        g_array_append_val (tmp_results, result);
-                }
+	if (search) {
+		return search->current_page_results->len;
+	} else {
+		return 0;
+	}
+}
 
-                ++i;
-        }
+gboolean
+pdf_document_find_get_result (EvDocumentFind *document_find,
+			      int             n_result,
+			      GdkRectangle   *rectangle)
+{
+	PdfDocument *pdf_document = PDF_DOCUMENT (document_find);
+	PdfDocumentSearch *search = pdf_document->search;
+	GdkRectangle r;
 
-        ev_document_find_found (EV_DOCUMENT_FIND (pdf_document),
-                                (EvFindResult*) tmp_results->data,
-                                tmp_results->len,
-                                pages_done / (double) n_pages);
+	if (search != NULL) {
+		r = g_array_index (search->current_page_results,
+				   GdkRectangle, n_result);
 
-        g_array_free (tmp_results, TRUE);
+		rectangle->x = r.x + pdf_document->page_x_offset;
+		rectangle->y = r.y + pdf_document->page_y_offset;
+		rectangle->width = r.width;
+		rectangle->height = r.height;
+
+		return TRUE;
+	} else {
+		return FALSE;
+	}
 }
 
 static void
@@ -461,7 +469,7 @@ pdf_document_search_page_changed (PdfDocumentSearch   *search)
 {
         PdfDocument *pdf_document = search->document;
         int current_page;
-        EvFindResult result;
+        GdkRectangle result;
         int xMin, yMin, xMax, yMax;
 
         current_page = pdf_document->page;
@@ -482,12 +490,10 @@ pdf_document_search_page_changed (PdfDocumentSearch   *search)
                                          gTrue, gTrue, // startAtTop, stopAtBottom
                                          gFalse, gFalse, // startAtLast, stopAtLast
                                          &xMin, &yMin, &xMax, &yMax)) {
-                result.page_num = pdf_document->page;
-
-                result.highlight_area.x = xMin;
-                result.highlight_area.y = yMin;
-                result.highlight_area.width = xMax - xMin;
-                result.highlight_area.height = yMax - yMin;
+                result.x = xMin;
+                result.y = yMin;
+                result.width = xMax - xMin;
+                result.height = yMax - yMin;
 
                 g_array_append_val (search->current_page_results, result);
                 /* Now find further results */
@@ -496,25 +502,14 @@ pdf_document_search_page_changed (PdfDocumentSearch   *search)
                                                     gFalse, gTrue,
                                                     gTrue, gFalse,
                                                     &xMin, &yMin, &xMax, &yMax)) {
-
-                        result.page_num = pdf_document->page;
-
-                        result.highlight_area.x = xMin;
-                        result.highlight_area.y = yMin;
-                        result.highlight_area.width = xMax - xMin;
-                        result.highlight_area.height = yMax - yMin;
+                        result.x = xMin;
+                        result.y = yMin;
+                        result.width = xMax - xMin;
+                        result.height = yMax - yMin;
 
                         g_array_append_val (search->current_page_results, result);
                 }
         }
-
-        /* needed for the initial current page since we don't search
-         * it in the idle
-         */
-        search->other_page_flags[current_page] =
-                search->current_page_results->len > 0;
-
-        pdf_document_search_emit_found (search);
 }
 
 static gboolean
@@ -522,7 +517,7 @@ pdf_document_search_idle_callback (void *data)
 {
         PdfDocumentSearch *search = (PdfDocumentSearch*) data;
         PdfDocument *pdf_document = search->document;
-        int n_pages;
+        int n_pages, changed_page;
         double xMin, yMin, xMax, yMax;
 
         /* Note that PDF page count is 1 through n_pages INCLUSIVE
@@ -534,10 +529,6 @@ pdf_document_search_idle_callback (void *data)
          */
         n_pages = ev_document_get_n_pages (EV_DOCUMENT (search->document));
 
-        if (search->search_page == search->start_page) {
-                goto end_search;
-        }
-
         if (search->output_dev == 0) {
                 /* First time through here... */
                 search->output_dev = new TextOutputDev (NULL, gTrue, gFalse, gFalse);
@@ -556,8 +547,12 @@ pdf_document_search_idle_callback (void *data)
                                           gFalse, gFalse, // startAtLast, stopAtLast
                                           &xMin, &yMin, &xMax, &yMax)) {
                 /* This page has results */
-                search->other_page_flags[search->search_page] = TRUE;
-        }
+                search->other_page_flags[search->search_page] = 1;
+        } else {
+		search->other_page_flags[search->search_page] = 0;
+	}
+
+	changed_page = search->start_page;
 
         search->search_page += 1;
         if (search->search_page > n_pages) {
@@ -565,12 +560,13 @@ pdf_document_search_idle_callback (void *data)
                 search->search_page = 1;
         }
 
-        /* We do this even if nothing was found, to update the percent complete */
-        pdf_document_search_emit_found (search);
+        if (search->search_page != search->start_page) {
+	        ev_document_find_changed (EV_DOCUMENT_FIND (pdf_document),
+					  changed_page);
+	        return TRUE;
+	}
 
-        return TRUE;
-
- end_search:
+end_search:
         /* We're done. */
         search->idle = 0; /* will return FALSE to remove */
         return FALSE;
@@ -583,7 +579,7 @@ pdf_document_find_begin (EvDocumentFind   *document,
 {
         PdfDocument *pdf_document = PDF_DOCUMENT (document);
         PdfDocumentSearch *search;
-        int n_pages;
+        int n_pages, i;
         gunichar *ucs4;
         glong ucs4_len;
 
@@ -618,13 +614,13 @@ pdf_document_find_begin (EvDocumentFind   *document,
 
         search->current_page_results = g_array_new (FALSE,
                                                     FALSE,
-                                                    sizeof (EvFindResult));
+                                                    sizeof (GdkRectangle));
         n_pages = ev_document_get_n_pages (EV_DOCUMENT (document));
 
-        /* This is an array of bool; with the first value ignored
-         * so we can index by the based-at-1 page numbers
-         */
-        search->other_page_flags = g_new0 (guchar, n_pages + 1);
+        search->other_page_flags = g_new0 (int, n_pages + 1);
+	for (i = 0; i <= n_pages; i++) {
+		search->other_page_flags[i] = -1;
+	}
 
         search->document = pdf_document;
 
@@ -637,9 +633,7 @@ pdf_document_find_begin (EvDocumentFind   *document,
         search->output_dev = 0;
 
         search->start_page = pdf_document->page;
-        search->search_page = search->start_page + 1;
-        if (search->search_page > n_pages)
-                search->search_page = 1;
+        search->search_page = search->start_page;
 
         search->current_page = -1;
 
@@ -1006,6 +1000,8 @@ pdf_document_get_title (PdfDocument *pdf_document)
 	char *title = NULL;
 	Object info;
 
+	if (pdf_document->doc == NULL)
+		return NULL;
 	pdf_document->doc->getDocInfo (&info);
 
 	if (info.isDict ()) {
@@ -1126,6 +1122,9 @@ static void
 pdf_document_find_iface_init (EvDocumentFindIface *iface)
 {
         iface->begin = pdf_document_find_begin;
+	iface->get_n_results = pdf_document_find_get_n_results;
+	iface->get_result = pdf_document_find_get_result;
+	iface->page_has_results = pdf_document_find_page_has_results;
         iface->cancel = pdf_document_find_cancel;
 }
 
