@@ -24,7 +24,6 @@
 
 #include "ev-document.h"
 #include "ev-link.h"
-#include "ev-history.h"
 
 G_BEGIN_DECLS
 
@@ -39,8 +38,6 @@ GType		ev_view_get_type	(void) G_GNUC_CONST;
 GtkWidget*	ev_view_new		(void);
 void		ev_view_set_document	(EvView     *view,
 			   		 EvDocument *document);
-void		ev_view_set_history	(EvView	    *view,
-					 EvHistory  *history);
 
 /* Clipboard */
 void		ev_view_copy		(EvView     *view);
@@ -64,6 +61,10 @@ void		ev_view_normal_size	(EvView     *view);
 void		ev_view_best_fit	(EvView     *view);
 void		ev_view_fit_width	(EvView     *view);
 
+/* Find */
+void            ev_view_find_next       (EvView     *view);
+void            ev_view_find_previous   (EvView     *view);
+
 /* Status */
 const char     *ev_view_get_status      (EvView     *view);
 const char     *ev_view_get_find_status (EvView     *view);
